{*****************************************************************************
* UNIT DESCRIPTION:  Form class for charting
* SPEC REF: (if used)
* DESIGN REF: (if used)
*
* Copyright (c) 2009 Schweitzer Engineering Laboratories, Pullman, Washington
******************************************************************************}
unit FormChartTimeSeries;

interface

uses
  Windows, Messages, SysUtils, StrUtils, DateUtils,
  Controls, Tabs, Forms, Dialogs, Buttons, ExtCtrls, Graphics,
  Classes, Contnrs, ToolWin, ComCtrls, StdCtrls, ClipBrd,
  TeeEdit, TeEngine, TeeTools, TeeProcs, TeePolar, Chart, Series,
  ClassTimeSeries, ClassTimeSeriesChart,
  ClassTimeSeriesChartSSI, ClassTimeSeriesChartHRM,
  ClassChartInput,
  ClassChartInputSSI,  ClassChartInputHRM,
  TimeRoutines, Types,
  TimeSeriesConstants, TeeScroB; //, ClassLoadProfileFile_010;
const
  INVALID_CHART_MSG = 'Invalid chart type.';
  CRLF = Chr(13) + Chr(10);     // carriage return line feed
  CR = Chr(13);     // carriage return
  TAB = Chr(9);
  INTERVAL_COLOR = clWebLightSlateGray;
  PARENTEVENT_COLOR = clYellow;
  SUBEVENT_COLOR = clRed;
  EVENTMARKER_COLOR = clYellow;
  DATETIME_WITH_SEC = 'MM/dd/yy hh:mm:ss';
  TIME_WITH_MS = 'hh:mm:ss.zzz';
  TIME_WITH_SEC = 'hh:mm:ss';
  SHGFP_TYPE_CURRENT         = 0;
  CSIDL_PERSONAL = $0005;  // Doc and settings > User > My Documents
  SPECTRAL_LIMIT = 3300;
  HARMONIC_LIMIT = 50;
  MARKERS_PER_EVENT = 4;
  MA_ARRAY_SIZE = 8;  // maximum array size for mag/angle evaluation
type
  TChartMenuEvent = procedure (form : TForm) of object;  // method pointer
  TArrayOfArrayOfDouble = array of array of double;
  TArrayOfDouble = array of double;
  TArrayOfInteger = array of integer;
  TArrayOfSingle = array of single;
  TArrayOfColor = array of TColor;

  TfrmChartTimeSeries = class(TForm)
    PanelChart: TPanel;
    SplitterEventForm: TSplitter;
    ScrollBoxChartControls: TScrollBox;
    ToolBar1: TToolBar;
    SpeedButtonPointer: TSpeedButton;
    SpeedButtonRotate: TSpeedButton;
    SpeedButtonTranslate: TSpeedButton;
    SpeedButtonZoom: TSpeedButton;
    SpeedButtonThickness: TSpeedButton;
    SpeedButton3D: TSpeedButton;
    SpeedButtonEdit: TSpeedButton;
    SpeedButtonCopy: TSpeedButton;
    SpeedButtonReset: TSpeedButton;
    SpeedButtonCursor: TSpeedButton;
    SpeedButtonDateTime: TSpeedButton;
    ChartHMIDisplay: TChart;
    ChartTool1: TCursorTool;
    ChartHMICBEMA: TChart;
    ChartTool2: TCursorTool;
    SpeedButtonLabel: TSpeedButton;
    SpeedButtonInformation: TSpeedButton;
    ComboBoxDateFormat: TComboBox;
    ButtonImport: TButton;
    SpeedButtonPrint: TSpeedButton;
    ChartScrollBarXAxis: TChartScrollBar;
    procedure ButtonImportClick(Sender: TObject);
    procedure ChartHMIDisplayGetNextAxisLabel(Sender: TChartAxis;
      LabelIndex: Integer; var LabelValue: Double; var Stop: Boolean);
    procedure ChartHMIDisplayMouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
  private
    FChartType              : TTimeSeriesChartType;   // chart type: ctLDP, ctSSI
    FChartSubType           : TTimeSeriesChartSubType;  // Sub-type of chart
    FChartDirty             : boolean;   // true if chart type changed since last display
    FUpdateEnabled          : boolean;  // true if display may be updated
    FDeviceType             : TTimeSeriesDeviceType;  // device type: dtSEL7XX...
    FChartTimeSeries        : TTimeSeriesChart;
    ChartSeries             : array of TCustomSeries;
    BarChartSeries          : array of TCustomBarSeries;
    PolarChartSeries        : array of TCustomPolarSeries;
    BarSeries               : TBarSeries;        // used for Harmonics
    PointSeries             : TPointSeries;  // used for LDP events
    FChartInputSettingsSSI  : TTimeSeriesInputSettingsSSI;
    FChartInputSettingsHRM  : TTimeSeriesInputSettingsHRM;
    FDateFormat             : string;  // date format string
    FMenuChart              : boolean;   // true if launched from tools menu
    FChartInstance          : integer;  // index identifying chart
    FChartAxisScale         : boolean;  // true if full harmonic axis, otherwise false
    FDefaultPath            : string;  // remember browse location during session
    FFileCount              : integer;
  public
    ChartMenuHandler : TChartMenuEvent; // delegate (method pointer)

    Constructor Create(aOwner: TComponent); reintroduce; overload;
    procedure SetChartType(aChartType : TTimeSeriesChartType);
    function  ReadChartData(aChartType : TTimeSeriesChartType;
                            fileReset, inputReset: boolean; filePath : string;
                            var ChartReset, ChartValid : boolean ) : boolean;
    procedure UpdateChart(reset : boolean);   overload;
    function  DisplayChartFile(filePath : string) : boolean;
    function  DisplayChart(aChartType : TTimeSeriesChartType; fileReset,inputReset: boolean; filePath : string; showChart : boolean ) : boolean; overload;
    procedure DisplayChart(found,chartReset,showChart,fileReset,inputReset,doCBEMA : boolean);  overload;
    procedure Display(reset : boolean);
    procedure InitChartMinMax(reset : boolean; var xMin,xMax,yMin,yMax : double);
    procedure UpdateChartMinMax(reset : boolean; xMin,xMax,yMin,yMax : double);
    procedure InitChartAttributes(channelCount : integer);
    procedure ChartHarmonicMagAnglePolar(index,channelCount : integer;
                myRecord : double;
                dataRecord : TTimeSeriesRecord);
    procedure ResetChartWindow;

    procedure FreeCharts();
    procedure AllocateChartArrays(
              channelCount : integer;
              harmSettings : TTimeSeriesInputSettingsHRM;
              var chartRec, eventRec, chartMarker, eventNumber : TArrayOfInteger;
              var eventColor : TArrayOfColor;
              var chartHeight, workArray : TArrayOfDouble;
              var xArray : TArrayOfDouble;
              var yArray : TArrayOfArrayOfDouble;
              var vbase, binsize : double;
              var doStatic : boolean);
    procedure FillChartArrays(channelCount, recordCount : integer;
              xArray : TArrayOfDouble; yArray : TArrayOfArrayOfDouble);
    procedure ChartLDPHarmonic(eventCount, channelCount : integer;
                doStatic : boolean; vbase, binSize : double;
                harmSettings : TTimeSeriesInputSettingsHRM;
                chartMarker, chartRec: TArrayOfInteger;
                chartHeight, workArray : TArrayOfDouble;
                xArray : TArrayOfDouble; yArray : TArrayOfArrayOfDouble);

    property ChartType : TTimeSeriesChartType read FChartType;
    property ChartSubType : TTimeSeriesChartSubType read FChartSubType write FChartSubType;
    property ChartDirty : boolean read FChartDirty;
    property UpdateEnabled : boolean read FUpdateEnabled;
    property DeviceType : TTimeSeriesDeviceType read FDeviceType;
    property ChartTimeSeries : TTimeSeriesChart read FChartTimeSeries;
    property ChartInputSettingsSSI : TTimeSeriesInputSettingsSSI read FChartInputSettingsSSI;
    property DateFormat : string read FDateFormat write FDateFormat;
    property MenuChart : boolean read FMenuChart write FMenuChart;
    property ChartInstance : integer read FChartInstance write FChartInstance;

  end;

implementation

Uses
  EditChar,
  TeeEditCha,
  Math;

{$R *.dfm}
{************************************************************************
* NAME:   Create
* FUNCTIONAL DESCRIPTION: class constructor
* INPUT PARAMETERS:  none
* OUTPUT PARAMETERS: none
* RETURN:            none
************************************************************************}
constructor TfrmChartTimeSeries.Create(aOwner: TComponent);
begin
  inherited Create(aOwner);
  FChartTimeSeries := TTimeSeriesChartHRM.Create(nil);
  FChartType := ctInvalid;    // nothing charted yet
  FChartSubType := cstInvalid;
  FChartDirty := true;
  FUpdateEnabled := true;
  FDateFormat := 'mm/dd/yyyy';

  self.HelpKeyword := 'Chart';

  FChartInstance := 0;
  FMenuChart := false;
  FChartAxisScale := false;
  FDefaultPath := '';
  FFileCount := 0;
end;

procedure TfrmChartTimeSeries.SetChartType(aChartType : TTimeSeriesChartType);
begin
  if(FChartType <> aChartType) then
  begin
    case aChartType of
      ctUndefined,ctHarmonic:
        begin
          // If input values not saved previously, create a default set for LDP
          if(FChartInputSettingsHRM = nil) then
          begin
            FChartInputSettingsHRM := TTimeSeriesInputSettingsHRM.Create;  // preserved input settings
            FChartInputSettingsHRM.ChartAxisDateTime := false;
          end;
          // InputSettings is a container for the input of the active chart
          if(FChartTimeSeries.EventData.InputSettings = nil) then
          begin
            FChartTimeSeries.EventData.InputSettings := FChartInputSettingsHRM;
          end;

          FChartTimeSeries.SaveChartActiveChannels;  // save settings in previous chart
          FChartTimeSeries.SaveChartActiveEvents;    // save SSI Channels before switching

          FreeAndNil(FChartTimeSeries);
          FChartTimeSeries := TTimeSeriesChartHRM.Create(nil);

          FChartTimeSeries.ChartDisplay := ChartHMIDisplay;
          FChartTimeSeries.EventData.DateFormat := DateFormat;
          // Make "InputSettings" a reference to the LDP input settings
          FChartTimeSeries.EventData.InputSettings := FChartInputSettingsHRM;
        end;

      ctSSI:
        begin
          // If input values not saved previously, create a default set for SSI
          if(FChartInputSettingsSSI = nil) then
          begin
            FChartInputSettingsSSI := TTimeSeriesInputSettingsSSI.Create;  // preserved input settings
          end;
          // InputSettings is a container for the input of the active chart
          if(FChartTimeSeries.EventData.InputSettings = nil) then
          begin
            FChartTimeSeries.EventData.InputSettings := FChartInputSettingsSSI;
          end;

          FChartTimeSeries.SaveChartActiveChannels;   // save settings in previous chart

          FreeAndNil(FChartTimeSeries);
          FChartTimeSeries := TTimeSeriesChartSSI.Create(nil);

          FChartTimeSeries.ChartDisplay := ChartHMIDisplay;
          FChartTimeSeries.EventData.DateFormat := DateFormat;
          // Make "InputSettings" a reference to the SSI input settings
          FChartTimeSeries.EventData.InputSettings := FChartInputSettingsSSI;
        end;
      ctInvalid:
      begin
        // do nothing
      end
      else
      begin
        raise Exception.Create(INVALID_CHART_MSG);
      end;
    end;
    FChartType := aChartType;   // now update the chart type
  end;
end;

procedure TfrmChartTimeSeries.ResetChartWindow;
var
  isHarmonic : boolean;
  harmonicInput : TTimeSeriesInputSettingsHRM;
begin
  isHarmonic := (FChartType = ctHarmonic);
  if(isHarmonic) then
  begin
    harmonicInput := FChartTimeSeries.EventData.InputSettings as TTimeSeriesInputSettingsHRM;
  end else
  begin
    harmonicInput := nil;
  end;

  ComboBoxDateFormat.Visible := true;

  if(FChartType = ctHarmonic) then
  begin

    FChartTimeSeries.EventData.InputSettings.ChartLabels := false;
    harmonicInput.HarmonicChartDisplayStart := 0;
    harmonicInput.HarmonicChartDisplayInterval := 5;
  end;

end;


function TfrmChartTimeSeries.ReadChartData(aChartType : TTimeSeriesChartType;
              fileReset, inputReset: boolean; filePath : string;
              var chartReset, chartValid : boolean ) : boolean;
const
  labelList : array[0..12] of string = ('Record','Date','Time','Ia','Ib','Ic','Ig','In','Va','Vb','Vc','Phase','Status');

var
  i,j,headerCount,dataCount : integer;
  dataRecord : TTimeSeriesRecord;
begin

  headerCount := 5;
  if(Pos('PIV',filePath) > 0) then
  begin
    SetChartType(ctHarmonic);
    dataCount := 7;
    ChartTimeSeries.EventData.ChartSubType := cstHRM_PIV;
  end else
  if(Pos('MAV',filePath) > 0) then
  begin
    SetChartType(ctHarmonic);
    dataCount := 9;
    ChartTimeSeries.EventData.ChartSubType := cstHRM_MAV;
  end else
  begin
    SetChartType(ctSSI);
    dataCount := 8;
    ChartTimeSeries.EventData.ChartSubType := cstInvalid;
  end;

  ChartTimeSeries.EventData.RawDataLabels.Clear();
  ChartTimeSeries.EventData.SetLengthHeaderColumns(headerCount);

  if(dataCount = 8) then
  begin
    ChartTimeSeries.EventData.HeaderColumns[0] := -1;
    ChartTimeSeries.EventData.HeaderColumns[1] := 1;
    ChartTimeSeries.EventData.HeaderColumns[2] := 2;
    ChartTimeSeries.EventData.HeaderColumns[3] := 11;
    ChartTimeSeries.EventData.HeaderColumns[4] := 12;
  end else
  begin
    for i := 0 to headerCount do
    begin
      ChartTimeSeries.EventData.HeaderColumns[i] := -1;
    end;
    ChartTimeSeries.EventData.HeaderColumns[0] := 0;
  end;

  ChartTimeSeries.EventData.SetLengthDataColumns(dataCount);
  for i := 0 to dataCount do
  begin
    ChartTimeSeries.EventData.DataColumns[i] := i+1;
    ChartTimeSeries.EventData.RawDataLabels.Add(labelList[i]);
  end;

  for i := 0 to 10  do
  begin
      dataRecord := TTimeSeriesRecord.Create;
      dataRecord.RecordNumber := i+1;

      if(dataCount = 8) then
      begin
        dataRecord.Phase := '...';
        dataRecord.Status := 'P';
        dataRecord.DateTime := Now;
        SetLength(dataRecord.Value,dataCount+1);
        dataRecord.Value[0] := i+2;
        for j := 1 to dataCount-1 do
        begin
          dataRecord.Value[j] := j;
        end;
      end else
      begin
        SetLength(dataRecord.Value,dataCount+1);
        dataRecord.Value[0] := i+2;
        for j := 1 to dataCount-1 do
        begin
          dataRecord.Value[j] := j;
        end;
      end;
      ChartTimeSeries.EventData.RawDataTable.Add(dataRecord);
  end;

  result := true;

end;

procedure TfrmChartTimeSeries.UpdateChart(reset : boolean);
begin
  FChartTimeSeries.SaveChartActiveChannels;   // remember which channels are active
  FChartTimeSeries.InitChartVariableCount;   // reset variable count if appropriate
  FChartTimeSeries.InitChartActiveChannelsDefault(false);   // re-allocate if necessary

  if(reset or ChartDirty) then   // chart type changed, reset charting controls
  begin
    ResetChartWindow;
  end;

    Display(reset);  // regenerate display based on currently checked events

end;

function TfrmChartTimeSeries.DisplayChartFile(filePath : string) : boolean;
var
  myChartType : TTimeSeriesChartType;
  myChartSubType : TTimeSeriesChartSubType;
  fileData : TStringList;
begin
  fileData := TStringList.Create;
  try
    if(Pos('PIV',filePath) > 0) then
    begin
      myChartType := ctHarmonic;
      myChartSubType := cstHRM_PIV;
    end else
    if(Pos('MAV',filePath) > 0) then
    begin
      myChartType := ctHarmonic;
      myChartSubType := cstHRM_MAV;
    end else
    begin
      myChartType := ctSSI;
      myChartSubType := cstInvalid;
    end;
    SetChartType(myChartType);
    FChartSubType := myChartSubType;
    ChartTimeSeries.EventData.ChartSubType := myChartSubType;
    result := DisplayChart(myChartType,true,true,filePath,true);
  finally
    fileData.Free;
  end;
end;

{************************************************************************
* NAME:   DisplayChart
* FUNCTIONAL DESCRIPTION:  Display Chart data found in file
* INPUT PARAMETERS:   aChartType : THMIChartType  type of chart to display
*                     fileReset : boolean  Initialize display of chart (true/false)
*                     inputReset : boolean  reset all user input parameters if true
*                     filename : string  is the original filename
*                     showChart : boolean  display the chart (true/false)
* OUTPUT PARAMETERS: none
* RETURN:             result : boolean  true if data could be parsed
************************************************************************}
function TfrmChartTimeSeries.DisplayChart(aChartType : TTimeSeriesChartType;
  fileReset,inputReset: boolean; filePath : string; showChart : boolean ) : boolean;
var
  chartReset, chartValid, found, doCBEMA : boolean;
  oldChartType : TTimeSeriesChartType;
  myInputSettings : TTimeSeriesInputSettingsSSI;
begin
  found := true;
  doCBEMA := false;
  // If display is already in progress, just return
  if(FUpdateEnabled) then
  begin
    oldChartType := FChartType;  // original type
    chartReset := true;

    if(fileReset and inputReset) then    // cancel date/time filter controls
    begin
      if(FChartType = ctSSI) then
      begin
        myInputSettings := (ChartTimeSeries.EventData.InputSettings as TTimeSeriesInputSettingsSSI);
        myInputSettings.BeginDateActive := false;
        myInputSettings.EndDateActive := false;
      end;
    end;

    try

      // Read chart data from file  (chartReset true if fileReset or change chart type)
      found := ReadChartData(aChartType, fileReset, inputReset, filePath, chartReset, chartValid);
      if(found) then
      begin
        if(not chartValid) then
        begin
//          ShowSimpleMessageBox('Error reading data into chart.');
        end else
        begin
          FUpdateEnabled := false;  // don't update display until finished generating it
          try
            DisplayChart(found,chartReset,showChart,fileReset,inputReset,doCBEMA);
            if(aChartType <> oldChartType) then
            begin
              ResetChartWindow();
              FChartTimeSeries.AutoScaleAxes(true);
            end;

            FUpdateEnabled := true;   // It is safe to update now

            FChartTimeSeries.ChartDisplay.Update();

          finally
            if(found or (not chartReset and (FChartTimeSeries.EventData.RawDataTable.Count > 1))) then
            begin
            // successful chart
            end else
            begin      // something went wrong
              FChartType := oldChartType;  // restore original chart type
            end;
          end;
        end;
      end;
    finally
      FUpdateEnabled := true;   // It is safe to update now
      FChartType := oldChartType;  // restore original chart type
    end;
  end;

  result := found;
end;

{************************************************************************
* NAME:   DisplayChart
* FUNCTIONAL DESCRIPTION:  Display Chart data already loaded into memory
* INPUT PARAMETERS:   found : boolean   data is in memory
*                     chartReset : boolean true fileReset or ChartType changed
*                     fileReset : boolean  true if new file imported
*                     inputReset : boolean true if reset user input options
*                     showChart : boolean  display the chart (true/false)
*                     doCBEMA : boolean true if CBEMA is default chart
* OUTPUT PARAMETERS: none
* RETURN:             none
************************************************************************}
procedure TfrmChartTimeSeries.DisplayChart(found,chartReset,showChart,fileReset,inputReset,doCBEMA : boolean);
var
  oldState : TWindowState;
begin
  if(found) then      // valid charting data found in file
  begin
    UpdateChart(inputReset);   // update the chart
  end;

  if(found or (not chartReset and (FChartTimeSeries.EventData.RawDataTable.Count > 1))) then
  begin
    if(showChart or Visible) then           // Display the chart window
    begin
      // NOTE: Delphi has unexpected behavior when chart window has been maximized.
      // This can be avoided by displaying in normal state, then switching to
      // original state.
      oldState := WindowState;
      WindowState := wsNormal;   // cancel window maximization
      Show;
      if(oldState <> wsNormal) then  // restore to original window size
      begin
        WindowState := oldState;
      end;
      if(FChartType = ctHarmonic) then
      begin
        ChartHMIDisplay.SetFocus;
      end;
    end;
  end;
end;

{************************************************************************
* NAME:   Display
* FUNCTIONAL DESCRIPTION:  Create the chart, assign attributes and insert data for display
* INPUT PARAMETERS:
*                    reset: boolean  true if regenerating display (rescale axis limits),
*                                    false if updating display (retain current axis limits)
* OUTPUT PARAMETERS: none
* RETURN:             none
************************************************************************}
procedure TfrmChartTimeSeries.Display(reset : boolean);
var
  eventCount,channelCount : integer;
  xMin,xMax,yMin,yMax : double;
  harmSettings : TTimeSeriesInputSettingsHRM;
  chartMarker,chartRec,eventRec,eventNumber : TArrayOfInteger;   // event number of charted events
  chartHeight,workArray : TArrayOfDouble;
  eventColor : TArrayofColor;
  vbase,binSize : double;
  table : TArrayOfArrayOfDouble;
  xArray : TArrayOfDouble;
  doStatic : boolean;
begin
  if(ChartTimeSeries.ChartVariableCount = 0) then exit;   // no data

  // default for harmonic chart only
  binSize := 5.0;

  channelCount := FChartTimeSeries.EventData.RawDataLabels.Count;  // number of channels to display

  eventCount := 0;
  if(FChartType = ctHarmonic) then
  begin
    harmSettings := FChartTimeSeries.EventData.InputSettings as TTimeSeriesInputSettingsHRM;
  end else
  begin
    harmSettings := nil;
  end;

  // Allocate dynamic arrays used for charting

  AllocateChartArrays(channelCount, harmSettings, chartRec, eventRec, chartMarker,
                      eventNumber, eventColor, chartHeight, workArray, xArray,
                      table, vbase, binsize, doStatic);

  // Initialize chart axis scale

  InitChartMinMax(reset,xMin,xMax,yMin,yMax);

  // Initialize chart attributes

  InitChartAttributes(channelCount);

  // Fill data into chart series

  if((FChartType = ctLDP) or (FChartType = ctHarmonic) or
     (FChartTimeSeries.EventData.IndexTable.Count = 0)) then     // no events, display all data
  begin
    ChartLDPHarmonic(eventCount, channelCount,
                doStatic, vbase, binSize, harmSettings,
                chartMarker, chartRec,
                chartHeight, workArray,
                xArray, table);
  end else
  begin
  end;

  // Update display attributes

  FChartTimeSeries.InitChartTitles;
  FChartTimeSeries.InitChartActiveChannels;
  UpdateChartMinMax(reset,xMin,xMax,yMin,yMax);
  FChartDirty := false;     // the chart display is now up-to-date
  try
    if(UpdateEnabled) then
    begin
      ChartHMIDisplay.Update();  // refresh display
    end;
  except
    // swallow exception
  end;
end;

{************************************************************************
* NAME:   AllocateChartArrays
* FUNCTIONAL DESCRIPTION: allocate working arrays for charting
* INPUT PARAMETERS: channelCount : integer number of channels (chart series)
*                   harmSettings : TTimeSeriesInputSettingsHRM harmonic chart settings
* OUTPUT PARAMETERS:
*                   chartRec : TArrayOfInteger  sequential record number
*                   eventRec : TArrayOfInteger  actual record number
*                   chartMarker : TArrayOfInteger  seq. record for marker
*                   eventNumber : TArrayOfInteger  event # of marker
*                   eventColor : TArrayOfColor   color of marker
*                   chartHeight : TArrayOfSingle height of marker
*                   workArray : TArrayOfSingle  used to revise magnitude/angle arrays
*                   xArray : array of double indep. variable record number
*                   yArray : array of array of double dep. variable data
*                   vbase : single
*                   binsize : single
*                   doStatic : boolean true if not using dynamic data arrays
* RETURN: none
************************************************************************}
procedure TfrmChartTimeSeries.AllocateChartArrays(
              channelCount : integer;
              harmSettings : TTimeSeriesInputSettingsHRM;
              var chartRec, eventRec, chartMarker, eventNumber : TArrayOfInteger;
              var eventColor : TArrayOfColor;
              var chartHeight, workArray : TArrayOfDouble;
              var xArray : TArrayOfDouble;
              var yArray : TArrayOfArrayOfDouble;
              var vbase, binsize : double;
              var doStatic : boolean);
var
  dataRecord     : TTimeSeriesRecord;
  minbin : double;
  index, count : integer;
begin
  // Make sure old chart data is freed before reallocating arrays
  FreeCharts();
  // Event marker information (possible 4 markers per event)
  count := FChartTimeSeries.EventData.indexTable.Count;
  if(count > 0) then
  begin
    SetLength(chartRec,count);  // sequential record number
    SetLength(eventRec,count);  // actual record number
    SetLength(chartMarker,count*MARKERS_PER_EVENT); // seq. record for marker
    SetLength(eventNumber,count*MARKERS_PER_EVENT); // event # of marker
    SetLength(eventColor,count*MARKERS_PER_EVENT); // color of marker
    SetLength(chartHeight,count*MARKERS_PER_EVENT); // height of marker
  end;

  if(FChartType = ctHarmonic) then
  begin
    if(FChartTimeSeries.EventData.RawDataTable.Count > 1) then
    begin
      dataRecord := (FChartTimeSeries.EventData.RawDataTable[FChartTimeSeries.EventData.RawDataTable.Count-1] as TTimeSeriesRecord);
      binSize := dataRecord.RecordValue;   // floating representation of (frequency)
      dataRecord := (FChartTimeSeries.EventData.RawDataTable[0] as TTimeSeriesRecord);
      minBin := dataRecord.RecordValue;     // smallest frequence bin
      binSize := (binSize - minBin)/FChartTimeSeries.EventData.RawDataTable.Count;
    end else
    begin
      Assert(false,'Not enough records for Harmonic chart');
    end;

    if((FChartType = ctHarmonic) and
         ((FChartSubType = cstHRM_MAI) or (FChartSubType = cstHRM_MAV))) then
    begin
      SetLength(PolarChartSeries,FChartTimeSeries.ChartVariableCount div 2);
    end else
    begin
      SetLength(BarChartSeries,FChartTimeSeries.ChartVariableCount);
    end;
    SetLength(ChartSeries,0);

    if((FChartType = ctHarmonic) and
         ((FChartSubType = cstHRM_MAI) or (FChartSubType = cstHRM_MAV))) then
    begin
      SetLength(workArray,MA_ARRAY_SIZE);  // used to revise magnitude/angle arrays
    end else
    if((FChartType = ctHarmonic) and (FChartSubType = cstHRM_INT) and
       (harmSettings.HarmonicChartType <> htSpectral)) then
    begin
      SetLength(workArray,channelCount);  // used as spectral work array
    end;

    if(FChartTimeSeries.EventData.SettingTable.Count > 0) then
    begin
      vbase := (FChartTimeSeries.EventData.SettingTable[0] as TTimeSeriesSettingRecord).SettingValue;
    end else
    begin
      vbase := 0.0;
    end;
  end else
  begin
    SetLength(BarChartSeries,0);
    SetLength(ChartSeries,FChartTimeSeries.ChartVariableCount);
    vbase := 0.0;
  end;

  // Set TRUE if harmonic chart data is not filled into a dynamic array
  doStatic := ((FChartType = ctHarmonic) and ((FChartSubType = cstHRM_INT) and
                             (harmSettings.HarmonicChartType <> htSpectral)) or
               ((FChartSubType = cstHRM_MAI) or (FChartSubType = cstHRM_MAV)));

  // Build array if not doStatic because it is faster for large datasets
  if(not doStatic) then
  begin
    SetLength(xArray, FChartTimeSeries.EventData.RawDataTable.Count);
    SetLength(yArray, channelCount);

    for index := 0 to channelCount - 1 do
    begin
      SetLength(yArray[index],FChartTimeSeries.EventData.RawDataTable.Count);
    end;
  end;
end;

{************************************************************************
* NAME:   FillChartArrays
* FUNCTIONAL DESCRIPTION: fill TeeChart with data using arrays for speed
* INPUT PARAMETERS: channelCount : integer number of channels (chart series)
*                   recordCount : integer number of records
*                   xArray : array of double indep. variable record number
*                   yArray : array of array of double dep. variable data
* OUTPUT PARAMETERS: none
* RETURN: none
************************************************************************}
procedure TfrmChartTimeSeries.FillChartArrays(channelCount, recordCount : integer;
              xArray : TArrayOfDouble; yArray : TArrayOfArrayOfDouble);
var
  index : integer;
begin
  // Send data to chart for each channel if using dynamic arrays for speed
  for index := 0 to channelCount - 1 do
  begin
    With FChartTimeSeries.ChartDisplay.Series[index].XValues do
    begin
      Value := TChartValues(xArray);
      Count := recordCount;
      Modified := true;
    end;
    With FChartTimeSeries.ChartDisplay.Series[index].YValues do
    begin
      Value := TChartValues(yArray[index]);
      Count := recordCount;
      Modified := true;
    end;
  end;
end;

{************************************************************************
* NAME:   ChartLDPHarmonic
* FUNCTIONAL DESCRIPTION: fill TeeChart arrays with data for LDP/Harmonic data
* INPUT PARAMETERS: eventCount : integer   active event count
*                   channelCount : integer number of channels (chart series)
*                   doStatic : boolean true if not using dynamic arrays
*                   vbase : single
*                   binSize : single   width of harmonic bin
*                   harmSettings : TTimeSeriesInputSettingsHRM
*                   chartMarker: TArrayOfInteger  seq. record for marker
*                   chartRec : TArrayOfInteger  sequential record number
*                   chartHeight : TArrayOfSingle   height of marker
*                   xArray : TArrayOfDouble      dynamic charting indep. var. arrays
*                   yArray : TArrayOfArrayOfDouble  dynamic charting dep var. arrays
* OUTPUT PARAMETERS: none
* RETURN: none
************************************************************************}
procedure TfrmChartTimeSeries.ChartLDPHarmonic(eventCount, channelCount : integer;
                doStatic : boolean; vbase, binSize : double;
                harmSettings : TTimeSeriesInputSettingsHRM;
                chartMarker, chartRec : TArrayOfInteger;
                chartHeight, workArray : TArrayOfDouble;
                xArray : TArrayOfDouble; yArray : TArrayOfArrayOfDouble);
const
  SKIP_RECORD = '(5)';
  TEST_RECORD = '(6)';
var
  index,j,rec : integer;
  myRecord : double;
  dataRecord        : TTimeSeriesRecord;
begin
  rec := 0;
  for index := 0 to FChartTimeSeries.EventData.RawDataTable.Count-1 do
  begin
    dataRecord := (FChartTimeSeries.EventData.RawDataTable[index] as TTimeSeriesRecord);
    if((dataRecord.Status = SKIP_RECORD) or (dataRecord.Status = TEST_RECORD)) then
    begin
      continue;  // don't display empty records
    end;
    if(FChartType = ctHarmonic) then
    begin
      myRecord := dataRecord.RecordValue;     // float representation
      if((FChartType = ctHarmonic) and
         ((FChartSubType = cstHRM_MAI) or (FChartSubType = cstHRM_MAV))) then
      begin
        ChartHarmonicMagAnglePolar(index,channelCount,myRecord,dataRecord);
      end else
      begin
          xArray[rec] := myRecord;
          for j := 0 to channelCount-1 do     // add every channel for this record to the chart
          begin
            yArray[j][rec] := dataRecord.Value[j];
          end;
          inc(rec);
      end;
    end else
    begin
      xArray[rec] := dataRecord.RecordNumber;
      for j := 0 to channelCount-1 do     // add every channel for this record to the chart
      begin
        yArray[j][rec] := dataRecord.Value[j];
      end;
      inc(rec);
    end;
  end;

  // Passing data to chart using arrays is faster for large data sets
  if(not doStatic) then
  begin
    FillChartArrays(channelCount, rec, xArray, yArray);
  end;

end;

{************************************************************************
* NAME:   ChartHarmonicMagAnglePolar
* FUNCTIONAL DESCRIPTION: Save updated dependent values for Harmonic Mag/Angle
* INPUT PARAMETERS: index : integer record number
*                   channelCount : integer  number of dependent values
*                   myRecord : double independent value
*                   dataRecord : TTimeSeriesRecord   data record
* OUTPUT PARAMETERS: none
* RETURN: none
************************************************************************}
procedure TfrmChartTimeSeries.ChartHarmonicMagAnglePolar(index,channelCount : integer;
              myRecord : double;
              dataRecord : TTimeSeriesRecord);
var
  j : integer;
  mag, angle : array of Double;
begin
  SetLength(mag,channelCount div 2);
  SetLength(angle,channelCount div 2);

  for j := 0 to channelCount-1 do     // add every channel for this record to the chart
  begin
    if((j mod 2) = 0) then    // magnitude
    begin
      mag[j div 2] := dataRecord.Value[j];
      angle[j div 2] := dataRecord.Value[j+1];
    end;
  end;

  for j := 0 to channelCount div 2-1 do     // add every channel for this record to the chart
  begin
    FChartTimeSeries.ChartDisplay.Series[j].AddXY(angle[j],mag[j]);
  end;
end;

{************************************************************************
* NAME:  InitChartMinMax
* FUNCTIONAL DESCRIPTION:  initialize chart axis scaling
* INPUT PARAMETERS: reset : boolean  true if discarding current scale information
* OUTPUT PARAMETERS:
*                   xMin : Double  minimum value of x-axis scale
*                   xMax : Double  maximum value of x-axis scale
*                   yMin : Double  minimum value of y-axis scale
*                   yMax : Double  maximum value of y-axis scale
* RETURN: none
************************************************************************}
procedure TfrmChartTimeSeries.InitChartMinMax(reset : boolean; var xMin,xMax,yMin,yMax : Double);
begin
    xMin := 0.0;
    xMax := 1.0;
    yMin := 0.0;
    yMax := 1.0;
  FChartTimeSeries.ChartDisplay.BottomAxis.Automatic := false;
  FChartTimeSeries.ChartDisplay.LeftAxis.Automatic := false;
end;

{************************************************************************
* NAME:    UpdateChartMinMax
* FUNCTIONAL DESCRIPTION:   Update chart axis scaling after chart is finished
* INPUT PARAMETERS: reset : boolean  true if discarding current scale information
*                   xMin : Double  minimum value of x-axis scale
*                   xMax : Double  maximum value of x-axis scale
*                   yMin : Double  minimum value of y-axis scale
*                   yMax : Double  maximum value of y-axis scale
* OUTPUT PARAMETERS: none
* RETURN: none
************************************************************************}
procedure TfrmChartTimeSeries.UpdateChartMinMax(reset : boolean; xMin,xMax,yMin,yMax : double);
const
  TOL = 1.0e-6;
begin
  if(reset) then
  begin
    FChartTimeSeries.AutoScaleAxes(false);   // use customized axis scaling
    if((FChartType = ctHarmonic) and (ChartScrollBarXAxis.Visible)) then
    begin
      ChartScrollBarXAxis.RecalcPosition;
    end;
  end else
  begin
    FChartTimeSeries.ChartDisplay.BottomAxis.Automatic := false;
    FChartTimeSeries.ChartDisplay.LeftAxis.Automatic := false;
    FChartTimeSeries.ChartDisplay.BottomAxis.Minimum := xMin;
    FChartTimeSeries.ChartDisplay.BottomAxis.Maximum := xMax;
    if((yMax - yMin) < TOL)  then    // Make sure there is some difference in the axis limits
    begin
      yMin := yMin - TOL;
      yMax := yMax + TOL;
    end;
    FChartTimeSeries.ChartDisplay.LeftAxis.Minimum := yMin;
    FChartTimeSeries.ChartDisplay.LeftAxis.Maximum := yMax;
  end;
end;

{************************************************************************
* NAME:   InitChartAttributes
* FUNCTIONAL DESCRIPTION:   Initialize attributes of the chart
* INPUT PARAMETERS: channelCount : integer   number of indep. variables
* OUTPUT PARAMETERS: none
* RETURN: none
************************************************************************}
procedure TfrmChartTimeSeries.InitChartAttributes(channelCount : integer);
const
  CBEMA_TITLES : array [0..1] of string = ('ITI Sag','ITI Swell');
  EVENT_TITLE = 'Event';
var
  j : integer;
begin

  // Allocate and initialize chart series

  for j := 0 to FChartTimeSeries.ChartVariableCount-1 do        // for each channel
  begin
    if(ChartType = ctHarmonic) then
    begin
      if((ChartSubType = cstHRM_MAI) or (ChartSubType = cstHRM_MAV)) then
      begin
        if(j < FChartTimeSeries.ChartVariableCount div 2) then  // combine mag/angle
        begin
          PolarChartSeries[j] := TPolarBarSeries.Create(self);    // create polar chart
        end;
      end else
      begin
        BarChartSeries[j] := TBarSeries.Create(self);    // create bar chart
      end;
    end else
    begin
      case FChartTimeSeries.EventData.InputSettings.ChartStyle of
        sctLine:
          begin
            ChartSeries[j] := TLineSeries.Create(self);    // create line chart
          end;
        sctArea:
          begin
            ChartSeries[j] := TAreaSeries.Create(self);   // create area chart
            TAreaSeries(ChartSeries[j]).Transparency := FChartTimeSeries.EventData.InputSettings.ChartTransparency;  // percent transparency
          end;
        else
        begin
          Assert(false,'Invalid chart style.');
        end;
      end;
    end;

    if(ChartType = ctHarmonic) then
    begin
      if((ChartSubType = cstHRM_MAI) or (ChartSubType = cstHRM_MAV)) then
      begin
        if(j < FChartTimeSeries.ChartVariableCount div 2) then  // combine mag/angle
        begin
          PolarChartSeries[j].Title := AnsiLeftStr(ChartTimeSeries.EventData.RawDataLabels[j*2],2);  // assign channel name to legend
          PolarChartSeries[j].Marks.Visible := ChartTimeSeries.EventData.InputSettings.ChartLabels;
          PolarChartSeries[j].CircleLabels := true;
          PolarChartSeries[j].ClockWiseLabels := false;
          FChartTimeSeries.ChartDisplay.AddSeries(PolarChartSeries[j]);           // add the new channel series to the chart
        end;
      end else
      begin
        BarChartSeries[j].Title := ChartTimeSeries.EventData.RawDataLabels[j];  // assign channel name to legend
        BarChartSeries[j].Marks.Visible := ChartTimeSeries.EventData.InputSettings.ChartLabels;
        FChartTimeSeries.ChartDisplay.AddSeries(BarChartSeries[j]);           // add the new channel series to the chart
      end;
    end else
    begin
      if(FChartTimeSeries.EventData.RawDataTable.Count > 50) then
      begin
        ChartSeries[j].AreaLinesPen.Visible := false;   // suppress grid lines
      end else
        ChartSeries[j].AreaLinesPen.Visible := true;

      if(j >= channelCount) then  // it is a CBEMA curve
      begin
        Assert(((j-channelCount) >= 0) and
                ((j-channelCount) < Length(CBEMA_TITLES)-1),
                'Array length mismatch');
        ChartSeries[j].Stairs := true;  // use step function style for CBEMA
        ChartSeries[j].Title := CBEMA_TITLES[j-channelCount];  // assign channel name to legend
      end else
      begin
        ChartSeries[j].Title := ChartTimeSeries.EventData.RawDataLabels[j];  // assign channel name to legend
      end;
      FChartTimeSeries.ChartDisplay.AddSeries(ChartSeries[j]);           // add the new channel series to the chart
    end;
  end;

end;

procedure TfrmChartTimeSeries.FreeCharts();
var
  count : integer;
begin
    try
      // There is a chronic problem related to freeing series in TeeChart.
      // This try/catch block is intended to minimize the impact of this problem.
      if(Assigned(ChartHMIDisplay)) then
      begin
        count := ChartHMIDisplay.SeriesCount();
        if(count > 0) then
        begin
          ChartHMIDisplay.FreeAllSeries();   // throw away previous chart data
        end;
      end;
    finally
      // nothing to do
    end;
end;


procedure TfrmChartTimeSeries.ButtonImportClick(Sender: TObject);
begin
  if(comboBoxDateFormat.Text = 'MDY') then
  begin
    FFileCount := FFileCount + 1;
  end;

  if( FFileCount mod 3 = 1) then      // CHANGING ORDER CHANGES SYMPTOM
  begin
    FDefaultPath := 'PIV';
  end else
  if( FFileCount mod 3 = 2) then
  begin
    FDefaultPath := 'MAV';
  end else
  begin
    FDefaultPath := 'SSI';
  end;
  DisplayChartFile(FDefaultPath);
end;

procedure TfrmChartTimeSeries.ChartHMIDisplayMouseMove(Sender: TObject;
  Shift: TShiftState; X, Y: Integer);
begin
end;

procedure TfrmChartTimeSeries.ChartHMIDisplayGetNextAxisLabel(Sender: TChartAxis;
  LabelIndex: Integer; var LabelValue: Double; var Stop: Boolean);
begin
  if(Sender = ChartHMIDisplay.BottomAxis) then
  begin
    FChartTimeSeries.ChartXDelta := Sender.CalcIncrement;
  end;
end;

initialization
  RegisterClass(TfrmChartTimeSeries);
end.

